import sys
import traceback
import copy
import io
import itertools
from collections import defaultdict

import matplotlib
matplotlib.use('Agg', warn=False)

import PIL
import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

from hits import utilities, interval, sam, sw

from . import experiment as experiment_module
from . import target_info as target_info_module
from . import layout as layout_module

class ReadDiagram():
    def __init__(self, 
                 alignments,
                 target_info,
                 R2_alignments=None,
                 ref_centric=False,
                 parsimonious=False,
                 zoom_in=None,
                 size_multiple=1,
                 draw_qualities=False,
                 draw_mismatches=True,
                 draw_polyA=False,
                 draw_sequence=False,
                 max_qual=41,
                 process_mappings=None,
                 detect_orientation=False,
                 label_layout=False,
                 highlight_SNPs=False,
                 reverse_complement=None,
                 label_left=False,
                 flip_donor=False,
                 flip_target=False,
                 read_label='sequencing read',
                 target_on_top=False,
                 features_on_alignments=True,
                 ax=None,
                 features_to_hide=None,
                 features_to_show=None,
                 refs_to_hide=None,
                 draw_edge_numbers=True,
                 hide_non_target_alignments=False,
                 default_color='grey',
                 color_overrides=None,
                 title=None,
                 mode='normal',
                 label_overrides=None,
                 split_at_indels=False,
                 only_target_and_donor=False,
                 force_left_aligned=False,
                 gap_between_read_pair=5,
                 width_per_unit=None,
                 arrow_width=None,
                 **kwargs):

        self.parsimonious = parsimonious

        self.target_info = target_info

        def clean_up_alignments(als):
            if als is None:
                als = []

            als = copy.deepcopy(als)
            if refs_to_hide is not None:
                als = [al for al in als if al.reference_name not in refs_to_hide]

            if self.parsimonious:
                als = interval.make_parsimonious(als)

            if split_at_indels:
                split_als = []
                for al in als:
                    if al.reference_name in [self.target_info.target, self.target_info.donor]:
                        split_at_dels = sam.split_at_deletions(al, 2)
                        split_at_ins = []
                        for al in split_at_dels:
                            split_at_ins.extend(sam.split_at_large_insertions(al, 2))

                        target_seq_bytes = self.target_info.reference_sequences[al.reference_name].encode()
                        extended = [sw.extend_alignment(al, target_seq_bytes) for al in split_at_ins]
                        split_als.extend(extended)

                    else:
                        split_als.append(al)

                als = split_als

            if only_target_and_donor:
                als = [al for al in als if al.reference_name in [self.target_info.target, self.target_info.donor]]

            return als

        if isinstance(alignments, dict):
            self.alignments = clean_up_alignments(alignments['R1'])
            self.R2_alignments = clean_up_alignments(alignments['R2'])
        else:
            self.alignments = clean_up_alignments(alignments)
            self.R2_alignments = None

        self.ref_centric = ref_centric
        self.target_on_top = target_on_top
        self.zoom_in = zoom_in
        self.size_multiple = size_multiple
        self.draw_qualities = draw_qualities
        self.draw_mismatches = draw_mismatches
        self.draw_polyA = draw_polyA
        self.draw_sequence = draw_sequence
        self.max_qual = max_qual
        self.process_mappings = process_mappings
        self.detect_orientation = detect_orientation
        self.label_layout = label_layout
        self.highlight_SNPs = highlight_SNPs
        self.reverse_complement = reverse_complement
        self.label_left = label_left
        self.flip_donor = flip_donor
        self.flip_target = flip_target
        self.read_label = read_label
        self.draw_edge_numbers = draw_edge_numbers
        self.features_on_alignments = features_on_alignments
        self.hide_non_target_alignments = hide_non_target_alignments
        self.default_color = default_color
        self.mode = mode
        self.force_left_aligned = force_left_aligned

        if self.mode == 'normal':
            self.font_sizes = {
                'read_label': 10,
                'ref_label': 10,
                'feature_label': 10,
                'number': 6,
                'sequence': 3.5 * self.size_multiple,
                'title': 16,
            }

            self.target_and_donor_y_gap = kwargs.get('target_and_donor_y_gap', 0.03)
            self.initial_alignment_y_offset = 5
            self.feature_line_width = 0.005
            self.gap_between_als = 0.003
            self.label_cut = False

            if self.label_left:
                self.label_x_offset = -30
            else:
                self.label_x_offset = 20

        elif self.mode == 'paper':
            self.font_sizes = {
                'read_label': 18,
                'ref_label': 18,
                'feature_label': 14,
                'number': 8,
                'sequence': 3.5 * self.size_multiple,
                'title': 16,
            }

            self.target_and_donor_y_gap = kwargs.get('target_and_donor_y_gap', 0.02)
            self.initial_alignment_y_offset = 5
            self.feature_line_width = 0.005
            self.gap_between_als = 0.003
            self.label_cut = True

            if self.label_left:
                self.label_x_offset = -30
            else:
                self.label_x_offset = 20

        elif self.mode == 'compact':
            self.font_sizes = {
                'read_label': 10,
                'ref_label': 16,
                'feature_label': 12,
                'number': 8,
                'sequence': 3.5 * self.size_multiple,
                'title': 22,
            }

            self.target_and_donor_y_gap = kwargs.get('target_and_donor_y_gap', 0.015)
            self.initial_alignment_y_offset = 3
            self.feature_line_width = 0.004
            self.gap_between_als = 0.003
            self.label_cut = False

            if self.label_left:
                self.label_x_offset = -5
            else:
                self.label_x_offset = 5
        
        if color_overrides is None:
            color_overrides = {}
        self.color_overrides = color_overrides
        self.title = title
        self.ax = ax

        if features_to_hide is None:
            self.features_to_hide = set()
        else:
            self.features_to_hide = features_to_hide
        
        self.features_to_show = features_to_show
        
        if len(self.alignments) > 0:
            self.query_length = self.alignments[0].query_length
            self.query_name = self.alignments[0].query_name
        else:
            self.query_length = 50
            self.query_name = None

        if self.R2_alignments is not None:
            self.R2_query_length = self.R2_alignments[0].query_length
        else:
            self.R2_query_length = None

        if width_per_unit is None:
            if self.query_length < 750:
                width_per_unit = 0.04
            elif 750 <= self.query_length < 2000:
                width_per_unit = 0.01
            else:
                width_per_unit = 0.005

        self.width_per_unit = width_per_unit

        self.height_per_unit = 40

        if self.ref_centric:
            self.arrow_linewidth = 3
        else:
            self.arrow_linewidth = 2

        if arrow_width is None:
            arrow_width = self.query_length * 0.007
        self.arrow_width = arrow_width
        self.arrow_height_over_width = self.width_per_unit / self.height_per_unit

        self.gap_between_read_pair = gap_between_read_pair

        self.text_y = -7

        self.cross_x = max(0.5, self.query_length * 0.002)
        self.cross_y = self.cross_x * self.width_per_unit / self.height_per_unit

        if self.ax is None:
            self.fig, self.ax = plt.subplots()
        else:
            self.fig = self.ax.figure
        
        if self.label_left:
            self.label_x = 0
            self.label_ha = 'right'
        else:
            self.label_x = 1
            self.label_ha = 'left'
        
        if self.reverse_complement is None:
            if self.detect_orientation and not all(al.is_unmapped for al in self.alignments):
                layout = layout_module.Layout(alignments, self.target_info)
                self.reverse_complement = (layout.strand == '-')

            else:
                self.reverse_complement = False

        if label_overrides is None:
            label_overrides = {}
        self.label_overrides = label_overrides

        self.feature_label_size = 10
        
        self.ref_name_to_color = defaultdict(lambda: default_color)
        self.ref_name_to_color[self.target_info.target] = 'C0'
        self.ref_name_to_color[self.target_info.donor] = 'C1'
        other_names = [n for n in self.target_info.reference_sequences if n not in [self.target_info.target, self.target_info.donor]]
        for i, name in enumerate(other_names):
            self.ref_name_to_color[name] = f'C{i % 8 + 2}'
        
        for name, color in self.color_overrides.items():
            self.ref_name_to_color[name] = color

        self.max_y = self.gap_between_als
        self.min_y = -self.gap_between_als 

        self.min_x = 0
        self.max_x = 1

        self.alignment_coordinates = defaultdict(list)

        self.plot_read()

        if self.ref_centric:
            self.draw_target_and_donor()

        self.update_size()

    def draw_read_arrows(self):
        ''' Draw black arrows that represent the sequencing read or read pair. '''
        arrow_kwargs = {
            'linewidth': self.arrow_linewidth,
            'color': 'black'
        }

        if self.R2_alignments is None:
            arrow_infos = [
                ([0, self.query_length - 1], self.reverse_complement),
            ]
        else:
            R2_start = self.query_length + self.gap_between_read_pair

            arrow_infos = [
                ([0, self.query_length - 1], False),
                ([R2_start, R2_start + self.R2_query_length - 1], True),
            ]

        for (x_start, x_end), reverse_complement in arrow_infos:
            if not reverse_complement:
                arrow_xs = [x_start, x_end, x_end - self.arrow_width]
            else:
                arrow_xs = [x_end, x_start, x_start + self.arrow_width]

            arrow_y = self.arrow_width * self.arrow_height_over_width
            arrow_ys = [0, 0, arrow_y]
            self.ax.plot(arrow_xs, arrow_ys, **arrow_kwargs)

        label_y_offset = 0

        self.ax.annotate(self.read_label,
                         xy=(self.label_x, 0),
                         xycoords=('axes fraction', 'data'),
                         xytext=(self.label_x_offset, label_y_offset),
                         textcoords='offset points',
                         color='black',
                         ha=self.label_ha,
                         va='center',
                         size=self.font_sizes['read_label'],
                        )

    def draw_alignments(self, alignments, is_R2=False):
        ax = self.ax

        # Copy before possibly fiddling with is_reverse below.
        alignments = copy.deepcopy([al for al in alignments if not al.is_unmapped])

        reverse_complement = self.reverse_complement or is_R2
        if is_R2:
            x_offset = self.query_length + self.gap_between_read_pair
        else:
            x_offset = 0

        # Ensure that target and donor are at the bottom followed by other references.
        reference_order = [self.target_info.target, self.target_info.donor]
        other_refs = sorted(set(al.reference_name for al in alignments if al.reference_name not in reference_order))
        reference_order += other_refs

        by_reference_name = defaultdict(list)
        for al in sorted(alignments, key=lambda al: (reference_order.index(al.reference_name), sam.query_interval(al))):
            by_reference_name[al.reference_name].append(al)
        
        if self.ref_centric:
            if self.target_on_top:
                rnames_below = [self.target_info.donor]
            else:
                rnames_below = [self.target_info.target]

            initial_offset = self.initial_alignment_y_offset

        else:
            rnames_below = []
            initial_offset = 1

        rnames_above = [n for n in by_reference_name if n not in rnames_below]

        offsets = {}
        for names, sign in [(rnames_below, -1), (rnames_above, 1)]:
            block_sizes = [initial_offset] + [len(by_reference_name[n]) + 1 for n in names]
            cumulative_block_sizes = np.cumsum(block_sizes)
            starts = sign * cumulative_block_sizes
            for name, start in zip(names, starts):
                offsets[name] = start
                if is_R2:
                    offsets[name] += sign

        for ref_name, ref_alignments in by_reference_name.items():
            if self.hide_non_target_alignments and ref_name != self.target_info.target:
                alpha_multiplier = 0
            else:
                alpha_multiplier = 1

            if reverse_complement:
                for alignment in ref_alignments:
                    alignment.is_reverse = not alignment.is_reverse

            offset = offsets[ref_name]
            color = self.ref_name_to_color[ref_name]

            average_y = (offset  + 0.5 * (len(ref_alignments) - 1)) * self.gap_between_als
            if (not self.ref_centric) or ref_name not in (self.target_info.target, self.target_info.donor):

                label = self.label_overrides.get(ref_name, ref_name)

                ax.annotate(label,
                            xy=(self.label_x, average_y),
                            xycoords=('axes fraction', 'data'),
                            xytext=(self.label_x_offset, 0),
                            textcoords='offset points',
                            color=color,
                            ha=self.label_ha,
                            va='center',
                            alpha=1 * alpha_multiplier,
                            size=self.font_sizes['ref_label'],
                        )
                        
            for i, alignment in enumerate(ref_alignments):
                start, end = sam.query_interval(alignment)

                def left_offset(x):
                    return x + x_offset - 0.5

                def right_offset(x):
                    return x + x_offset + 0.5

                def middle_offset(x):
                    return x + x_offset

                strand = sam.get_strand(alignment)
                y = (offset + i * np.sign(offset)) * self.gap_between_als
                
                # Annotate the ends of alignments with reference position numbers and vertical lines.
                for x, which, offset_function in ((start, 'start', left_offset), (end, 'end', right_offset)):
                    final_x = offset_function(x)

                    if (which == 'start' and strand == '+') or (which == 'end' and strand == '-'):
                        r = alignment.reference_start
                    else:
                        r = alignment.reference_end - 1

                    ax.plot([final_x, final_x], [0, y], color=color, alpha=0.3 * alpha_multiplier)

                    if which == 'start':
                        kwargs = {'ha': 'right', 'xytext': (-2, 0)}
                    else:
                        kwargs = {'ha': 'left', 'xytext': (2, 0)}

                    if self.draw_edge_numbers or ref_name not in (self.target_info.target, self.target_info.donor):
                        ax.annotate('{0:,}'.format(r),
                                    xy=(final_x, y),
                                    xycoords='data',
                                    textcoords='offset points',
                                    color=color,
                                    va='center',
                                    size=self.font_sizes['number'],
                                    alpha=1 * alpha_multiplier,
                                    **kwargs)

                if self.draw_mismatches:
                    mismatches = layout_module.get_mismatch_info(alignment, self.target_info)
                    for read_p, read_b, ref_p, ref_b, q in mismatches:
                        if q < self.max_qual * 0.75:
                            alpha = 0.25
                        else:
                            alpha = 0.85

                        cross_kwargs = dict(zorder=10, color='black', alpha=alpha * alpha_multiplier)
                        cross_ys = [y - self.cross_y, y + self.cross_y]

                        read_x = middle_offset(read_p)

                        ax.plot([read_x - self.cross_x, read_x + self.cross_x], cross_ys, **cross_kwargs)
                        ax.plot([read_x + self.cross_x, read_x - self.cross_x], cross_ys, **cross_kwargs)

                # Draw the alignment, with downward dimples at insertions and upward loops at deletions.
                xs = [left_offset(start)]
                ys = [y]
                indels = sorted(layout_module.get_indel_info(alignment), key=lambda t: t[1][0])
                for kind, info in indels:
                    if kind == 'deletion' or kind == 'splicing':
                        centered_at, length = info

                        if kind == 'deletion':
                            max_length = 100
                            label = str(length)
                        else:
                            max_length = 10
                            label = 'splicing'

                        # Cap how wide the loop can be.
                        capped_length = min(max_length, length)
                        
                        if length <= 1:
                            height = 0.0015
                            indel_xs = [centered_at, centered_at, centered_at]
                            indel_ys = [y, y + height, y]
                        else:
                            width = self.query_length * 0.001
                            height = 0.003

                            indel_xs = [
                                centered_at - width,
                                centered_at - 0.5 * capped_length,
                                centered_at + 0.5 * capped_length,
                                centered_at + width,
                            ]
                            indel_ys = [y, y + height, y + height, y]

                            ax.annotate(label,
                                        xy=(middle_offset(centered_at), y + height),
                                        xytext=(0, 1),
                                        textcoords='offset points',
                                        ha='center',
                                        va='bottom',
                                        size=6,
                                        alpha=1 * alpha_multiplier,
                                    )

                    elif kind == 'insertion':
                        starts_at, ends_at = info
                        centered_at = np.mean([starts_at, ends_at])
                        length = ends_at - starts_at + 1

                        min_height = 0.0015
                        height = min_height * min(length**0.5, 3)
                        if length > 1:
                            ax.annotate(str(length),
                                        xy=(centered_at, y - height),
                                        xytext=(0, -1),
                                        textcoords='offset points',
                                        ha='center',
                                        va='top',
                                        size=6,
                                        alpha=1 * alpha_multiplier,
                                    )
                        indel_xs = [starts_at - 0.5, centered_at, ends_at + 0.5]
                        indel_ys = [y, y - height, y]
                        
                    xs.extend([middle_offset(x) for x in indel_xs])
                    ys.extend(indel_ys)
                    
                xs.append(right_offset(end))
                ys.append(y)

                ref_ps = (alignment.reference_start, alignment.reference_end - 1)
                if alignment.is_reverse:
                    ref_ps = ref_ps[::-1]

                coordinates = [
                    (middle_offset(start), middle_offset(end)),
                    ref_ps,
                    y,
                    sam.get_strand(alignment),
                ]
                self.alignment_coordinates[ref_name].append(coordinates)
                
                self.max_y = max(self.max_y, max(ys))
                self.min_y = min(self.min_y, min(ys))
                
                kwargs = {
                    'color': color,
                    'linewidth': 1.5,
                    'alpha': 1 * alpha_multiplier,
                }

                ax.plot(xs, ys, **kwargs)

                length = end - start

                capped_arrow_width = min(self.arrow_width, length * 0.3)
                
                if strand == '+':
                    arrow_xs = [right_offset(end), right_offset(end - capped_arrow_width)]
                    arrow_ys = [y, y + capped_arrow_width * self.arrow_height_over_width]
                else:
                    arrow_xs = [left_offset(start), left_offset(start + capped_arrow_width)]
                    arrow_ys = [y, y - capped_arrow_width * self.arrow_height_over_width]
                    
                draw_arrow = True
                if self.zoom_in is not None:
                    if not all(self.min_x <= x <= self.max_x for x in arrow_xs):
                        draw_arrow = False

                if draw_arrow:
                    ax.plot(arrow_xs, arrow_ys, clip_on=False, **kwargs)

                features = copy.deepcopy(self.target_info.features)

                if self.features_to_show is not None:
                    features_to_show = self.features_to_show
                else:
                    features_to_show = [
                        (r_name, f_name) for r_name, f_name in features
                        if r_name == ref_name and
                        'edge' not in f_name and
                        'SNP' not in f_name and 
                        features[r_name, f_name].feature != 'sgRNA'
                    ]

                    features_to_show.extend([(ref_name, f_name) for f_name in self.target_info.sgRNA_features])

                q_to_r = {
                    sam.true_query_position(q, alignment): r
                    for q, r in alignment.aligned_pairs
                    if r is not None and q is not None
                }

                if self.highlight_SNPs:
                    if ref_name == self.target_info.donor:
                        SNVs = self.target_info.donor_SNVs['donor']
                    elif ref_name == self.target_info.target:
                        SNVs = self.target_info.donor_SNVs['target']
                    else:
                        SNVs = {}

                    for SNV_name, SNV_info in SNVs.items():
                        SNV_r = SNV_info['position']
                        qs = [q for q, r in q_to_r.items() if r == SNV_r]
                        if len(qs) != 1:
                            continue

                        q = qs[0]

                        left_x = q - 0.5
                        right_x = q + 0.5
                        bottom_y = y - (self.cross_y * 3)
                        top_y = y + (self.cross_y * 3)
                        path_xs = [left_x, right_x, right_x, left_x]
                        path_ys = [bottom_y, bottom_y, top_y, top_y]
                        path = np.array([path_xs, path_ys]).T
                        patch = plt.Polygon(path, color='black', alpha=0.2, linewidth=0)
                        ax.add_patch(patch)
                        
                for feature_reference, feature_name in features_to_show:
                    if ref_name != feature_reference:
                        continue

                    if (feature_reference, feature_name) not in features:
                        continue

                    if feature_name in self.features_to_hide:
                        continue

                    feature = features[feature_reference, feature_name]
                    feature_color = feature.attribute.get('color', 'grey')
                    
                    qs = [q for q, r in q_to_r.items() if feature.start <= r <= feature.end]
                    if not qs:
                        continue

                    query_extent = [min(qs), max(qs)]
                    
                    rs = [feature.start, feature.end]
                    if strand == '-':
                        rs = rs[::-1]

                    if np.sign(offset) == 1:
                        va = 'bottom'
                        text_y = 1
                    else:
                        va = 'top'
                        text_y = -1

                    if not self.ref_centric:
                        for ha, q, r in zip(['left', 'right'], query_extent, rs):
                            nts_missing = abs(q_to_r[q] - r)
                            if nts_missing != 0 and query_extent[1] - query_extent[0] > 20:
                                ax.annotate(str(nts_missing),
                                            xy=(q, 0),
                                            ha=ha,
                                            xytext=(3 if ha == 'left' else -3, text_y),
                                            textcoords='offset points',
                                            size=6,
                                            va=va,
                                        )
                        
                        if query_extent[1] - query_extent[0] > 18 or feature.attribute['ID'] == self.target_info.sgRNA:
                            label = feature.attribute['ID']

                            ax.annotate(label,
                                        xy=(np.mean(query_extent), 0),
                                        xycoords='data',
                                        xytext=(0, self.text_y),
                                        textcoords='offset points',
                                        va='top',
                                        ha='center',
                                        color=feature_color,
                                        size=10,
                                        weight='bold',
                                    )

                    if self.features_on_alignments:
                        xs = [min(qs) - 0.5 + x_offset, max(qs) + 0.5 + x_offset]
                        ax.fill_between(xs, [y] * 2, [0] * 2, color=feature_color, alpha=0.7)
                        

    def plot_read(self):
        ax = self.ax

        if (not self.alignments) or (self.alignments[0].query_sequence is None):
            return self.fig

        if self.zoom_in is not None:
            self.min_x = self.zoom_in[0]
            self.max_x = self.zoom_in[1]
        else:
            if self.R2_alignments is None:
                total_query_length = self.query_length
            else:
                total_query_length = self.query_length + self.R2_query_length + self.gap_between_read_pair

            self.min_x = -0.02 * total_query_length
            self.max_x = 1.02 * total_query_length
            
        self.draw_read_arrows()

        self.draw_alignments(self.alignments)

        if self.R2_alignments is not None:
            self.draw_alignments(self.R2_alignments, is_R2=True)

        if self.title is None:
            if self.label_layout:
                layout = layout_module.Layout(self.alignments, self.target_info)
                cat, subcat, details = layout.categorize()
                title = '{}\n{}, {}, {}'.format(self.query_name, cat, subcat, details)
            else:
                title = self.query_name
        else:
            title = self.title

        ax.set_title(title, y=1.02, size=self.font_sizes['title'])
            
        ax.set_ylim(1.1 * self.min_y, 1.1 * self.max_y)
        ax.set_xlim(self.min_x, self.max_x)
        ax.set_yticks([])
        
        ax.spines['bottom'].set_position(('data', 0))
        ax.spines['bottom'].set_alpha(0.1)
        for edge in 'left', 'top', 'right':
            ax.spines[edge].set_color('none')
            
        # Set how far tick labels should be from axes
        if self.draw_sequence:
            ax.tick_params(length=0, pad=self.font_sizes['sequence'] * 1.5)
        elif not self.ref_centric:
            ax.tick_params(pad=14)

        if self.draw_qualities:
            def quals_to_ys(quals):
                return (np.array(quals) + 5) / self.max_qual * 2 * self.gap_between_als

            qual_ys = quals_to_ys(self.alignments[0].get_forward_qualities())

            ax.plot(qual_ys, color='black', alpha=0.5)

            if self.R2_alignments is not None:
                qual_ys = quals_to_ys(self.R2_alignments[0].get_forward_qualities())

                x_start = self.query_length + self.gap_between_read_pair
                xs = x_start + np.arange(self.R2_query_length)[::-1]

                ax.plot(xs, qual_ys, color='black', alpha=0.5)

        if self.draw_polyA:
            seq = self.alignments[0].get_forward_sequence()
            for b, color in [('A', 'red'), ('G', 'brown')]:
                locations = utilities.homopolymer_lengths(seq, b)
                for start, length in locations:
                    if length > 10:
                        ax.fill_between([start, start + length - 1], [self.max_y + self.arrow_height] * 2, [0] * 2, color=color, alpha=0.2)
                        
                        ax.annotate('poly{}'.format(b),
                                    xy=(start + length / 2, 0),
                                    xycoords='data',
                                    xytext=(0, self.text_y),
                                    textcoords='offset points',
                                    va='top',
                                    ha='center',
                                    color=color,
                                    alpha=0.4,
                                    size=10,
                                    weight='bold',
                                )
                        
        if self.draw_sequence:
            seq = self.alignments[0].get_forward_sequence()
            seq_kwargs = dict(family='monospace',
                              size=self.font_sizes['sequence'],
                              ha='center',
                              textcoords='offset points',
                              va='top',
                              xytext=(0, -2),
                             )

            for x, b in enumerate(seq):
                if self.min_x <= x <= self.max_x:
                    ax.annotate(b, xy=(x, 0), **seq_kwargs)
            
            if self.R2_alignments is not None:
                seq = utilities.reverse_complement(self.R2_alignments[0].get_forward_sequence())

                x_start = self.query_length + self.gap_between_read_pair
                for x_offset, b in enumerate(seq):
                    x = x_start + x_offset
                    if self.min_x <= x <= self.max_x:
                        ax.annotate(b, xy=(x, 0), **seq_kwargs)
            
        return self.fig

    def draw_target_and_donor(self):
        if len(self.alignments) == 0:
            return

        def adjust_edges(xs):
            xs = list(xs)

            if xs[0] < xs[1]:
                xs[0] -= 0.5
                xs[1] += 0.5
            else:
                xs[0] += 0.5
                xs[1] -= 0.5

            return xs

        ti = self.target_info
        ref_line_width = 0.001
        
        if self.target_on_top:
            target_y = self.max_y + self.target_and_donor_y_gap
            donor_y = self.min_y - self.target_and_donor_y_gap
        else:
            target_y = self.min_y - self.target_and_donor_y_gap
            donor_y = self.max_y + self.target_and_donor_y_gap

        params = []

        if len(self.alignment_coordinates[ti.target]) > 0:
            params.append((ti.target, min(ti.cut_afters.values()), target_y, self.flip_target))

        if len(self.alignment_coordinates[ti.donor]) > 0:
            if (ti.donor, ti.donor_specific) in ti.features:
                donor_specific_feature = ti.features[ti.donor, ti.donor_specific]
                middle = np.mean([donor_specific_feature.start, donor_specific_feature.end])
            else:
                middle = len(ti.donor_sequence) / 2

            params.append((ti.donor, middle, donor_y, self.flip_donor))

        for ref_name, center_p, ref_y, flip in params:
            color = self.ref_name_to_color[ref_name]

            # To establish a mapping between reference position and x coordinate,
            # pick anchor points on the ref and read that will line up with each other. 
            if self.force_left_aligned or (len(self.alignment_coordinates[ref_name]) == 1):
                xs, ps, y, strand = self.alignment_coordinates[ref_name][0]

                anchor_ref = ps[0]
                if (strand == '+' and not flip) or (strand == '-' and flip):
                    anchor_read = xs[0]
                else:
                    anchor_read = xs[1]
            else:
                anchor_ref = center_p

                total_query_length = self.query_length
                if self.R2_query_length is not None:
                    total_query_length += self.gap_between_read_pair + self.R2_query_length

                anchor_read = total_query_length // 2

            # With these anchors picked, define the mapping and its inverse.
            if flip:
                ref_p_to_x = lambda p: anchor_read - (p - anchor_ref)
                x_to_ref_p = lambda x: (anchor_read - x) + anchor_ref
            else:
                ref_p_to_x = lambda p: (p - anchor_ref) + anchor_read
                x_to_ref_p = lambda x: (x - anchor_read) + anchor_ref

            ref_edge = len(ti.reference_sequences[ref_name]) - 1

            # ref_start and ref_end are the smallest and largest ref positions
            # that get plottted. Initially, set these to the inverse image of
            # the edges of the current x lims.
            if flip:
                left, right = self.max_x, self.min_x
            else:
                left, right = self.min_x, self.max_x

            ref_start = max(0, x_to_ref_p(left))
            ref_end = min(ref_edge, x_to_ref_p(right))

            ref_al_min = np.inf
            ref_al_max = -np.inf

            for xs, ps, y, strand in self.alignment_coordinates[ref_name]:
                ref_xs = [ref_p_to_x(p) for p in ps]
                ref_al_min = min(ref_al_min, min(ps))
                ref_al_max = max(ref_al_max, max(ps))

                xs = adjust_edges(xs)
                ref_xs = adjust_edges(ref_xs)

                parallelogram_alpha = 0.05
                # Shade parallelograms between alignments and reference.
                if ref_y < 0:
                    ref_border_y = ref_y + ref_line_width
                else:
                    ref_border_y = ref_y - ref_line_width

                self.ax.fill_betweenx([y, ref_border_y], [xs[0], ref_xs[0]], [xs[1], ref_xs[1]], color=color, alpha=parallelogram_alpha)

                # Draw lines connecting alignment edges to reference.
                for x, ref_x in zip(xs, ref_xs):
                    self.ax.plot([x, ref_x], [y, ref_border_y], color=color, alpha=0.3)

            self.min_y = min(self.min_y, ref_y)
            self.max_y = max(self.max_y, ref_y)

            if ref_al_min <= ref_start:
                ref_start = max(0, ref_al_min - 10)

            if ref_al_max >= ref_end:
                ref_end = min(ref_edge, ref_al_max + 10)

            if ref_name == ti.donor and ref_edge < 3000:
                ref_end = ref_edge

            if ref_name == ti.target:
                fade_left = True
                fade_right = True
            else:
                fade_left = (ref_start > 0)
                fade_right = (ref_end < ref_edge)

            new_left = ref_p_to_x(ref_start)
            new_right = ref_p_to_x(ref_end)

            leftmost_aligned_x = ref_p_to_x(ref_al_min)
            rightmost_aligned_x = ref_p_to_x(ref_al_max)

            if flip:
                new_left, new_right = new_right, new_left
                fade_left, fade_right = fade_right, fade_left
                leftmost_aligned_x, rightmost_aligned_x = rightmost_aligned_x, leftmost_aligned_x

            self.min_x = min(self.min_x, new_left)
            self.max_x = max(self.max_x, new_right)

            # If an alignment goes right up to the edge, don't fade.
            if leftmost_aligned_x <= self.min_x + 0.1 * (self.max_x - self.min_x):
                fade_left = False

            if rightmost_aligned_x >= self.min_x + 0.9 * (self.max_x - self.min_x):
                fade_right = False

            # If a plasmid donor is being drawn, fade.
            #if ref_name == ti.donor and ti.donor_type == 'plasmid':
            #    fade_left = True
            #    fade_right = True

            self.ax.set_xlim(self.min_x, self.max_x)

            # Draw the actual reference.
            if ref_name == ti.target:
                # Always extend the target all the way to the edges.
                ref_xs = [self.min_x, self.max_x]
            else:
                ref_xs = adjust_edges([ref_p_to_x(ref_start), ref_p_to_x(ref_end)])

            rgba = matplotlib.colors.to_rgba(color)
            image = np.expand_dims(np.array([rgba]*1000), 0)

            if fade_left:
                left_alpha = 0
            else:
                left_alpha = 1

            if fade_right:
                right_alpha = 0
            else:
                right_alpha = 1

            image[:, :, 3] = np.concatenate([np.linspace(left_alpha, 1, 150), [1] * 700, np.linspace(1, right_alpha, 150)])
            self.ax.imshow(image, extent=(ref_xs[0], ref_xs[1], ref_y - ref_line_width, ref_y + ref_line_width), aspect='auto')

            if self.features_to_show is not None:
                features_to_show = [(r_name, f_name) for r_name, f_name in self.features_to_show
                                    if r_name == ref_name
                                   ]
            else:
                features_to_show = [(r_name, f_name) for r_name, f_name in ti.features
                                    if r_name == ref_name
                                    and 'edge' not in f_name
                                    and 'SNP' not in f_name
                                    and 'sgRNA' not in f_name
                                   ]

                for sgRNA in self.target_info.sgRNAs:
                    features_to_show.append((self.target_info.target, sgRNA))

            # Draw features.
            for feature_reference, feature_name in features_to_show:
                if feature_reference != ref_name:
                    continue

                if feature_name is None or feature_name in self.features_to_hide:
                    continue

                feature = ti.features.get((feature_reference, feature_name))
                if feature is None:
                    continue
                feature_color = feature.attribute.get('color', 'grey')
                
                xs = adjust_edges([ref_p_to_x(p) for p in [feature.start, feature.end]])
                    
                start = ref_y + np.sign(ref_y) * ref_line_width
                end = start + np.sign(ref_y) * self.feature_line_width

                bottom = min(start, end)
                top = max(start, end)
                self.min_y = min(bottom, self.min_y)
                self.max_y = max(top, self.max_y)

                # + 0.5 in this check to account for adjust_edges
                if min(xs) >= self.min_x and max(xs) <= self.max_x + 0.5:
                    self.ax.fill_between(xs, [start] * 2, [end] * 2, color=feature_color, alpha=0.7, linewidth=0)

                    name = feature.attribute['ID']
                    label = self.label_overrides.get(name, name)

                    self.ax.annotate(label,
                                     xy=(np.mean(xs), end),
                                     xycoords='data',
                                     xytext=(0, 5 * np.sign(ref_y)),
                                     textcoords='offset points',
                                     va='top' if ref_y < 0 else 'bottom',
                                     ha='center',
                                     color=feature_color,
                                     size=self.font_sizes['feature_label'],
                                     weight='bold',
                                    )

            # Draw target and donor names next to diagrams.
            label = self.label_overrides.get(ref_name, ref_name)

            self.ax.annotate(label,
                             xy=(self.label_x, ref_y),
                             xycoords=('axes fraction', 'data'),
                             xytext=(self.label_x_offset, 0),
                             textcoords='offset points',
                             color=color,
                             ha=self.label_ha,
                             va='center',
                             size=self.font_sizes['ref_label'],
                            )

            # Draw the cut site(s).
            if ref_name == ti.target:
                for cut_after in ti.cut_afters.values():
                    cut_after_x = ref_p_to_x(cut_after)
                    cut_y_bottom = ref_y + np.sign(ref_y) * self.feature_line_width * 1.5
                    cut_y_top = ref_y - np.sign(ref_y) * self.feature_line_width
                    self.ax.plot([cut_after_x, cut_after_x], [cut_y_bottom, cut_y_top], '--', color='black')

                    if self.label_cut:
                        self.ax.annotate('cut site',
                                        xy=(cut_after_x, cut_y_bottom),
                                        xycoords='data',
                                        xytext=(0, 5 * np.sign(ref_y)),
                                        textcoords='offset points',
                                        color='black',
                                        ha='center',
                                        va='top' if ref_y < 0 else 'bottom',
                                        size=self.font_sizes['ref_label'],
                                        )
            
        self.ax.set_ylim(self.min_y - 0.1 * self.height, self.max_y + 0.1 * self.height)

    @property
    def height(self):
        return self.max_y - self.min_y
    
    @property
    def width(self):
        return self.max_x - self.min_x

    def update_size(self):
        #fig_width = 0.04 * (self.width + 50) * self.size_multiple
        #fig_height = 40 * self.height * self.size_multiple

        fig_width = self.width_per_unit * max(self.width, 50) * self.size_multiple
        fig_height = self.height_per_unit * 1.2 * self.height * self.size_multiple

        self.fig.set_size_inches((fig_width, fig_height))

def make_stacked_Image(diagrams, titles=None, **kwargs):
    if titles is None or titles == '':
        titles = itertools.repeat(titles)

    ims = []

    for diagram, title in zip(diagrams, titles):
        if title is not None:
            diagram.fig.axes[0].set_title(title)
        
        with io.BytesIO() as buffer:
            diagram.fig.savefig(buffer, format='png', bbox_inches='tight')
            im = PIL.Image.open(buffer)
            im.load()
            ims.append(im)

        plt.close(diagram.fig)
        
    if not ims:
        return None

    total_height = sum(im.height for im in ims)
    max_width = max(im.width for im in ims)

    stacked_im = PIL.Image.new('RGBA', size=(max_width, total_height), color='white')
    y_start = 0
    for im in ims:
        stacked_im.paste(im, (max_width - im.width, y_start))
        y_start += im.height

    return stacked_im



import functools
import io
import base64
import os
from pathlib import Path

import pandas as pd
import nbconvert
import nbformat.v4 as nbf
import PIL

from . import experiment
from . import svg

totals_all_row_label = (' ', 'Total reads')

totals_relevant_row_label = (' ', 'Total relevant reads')

def load_counts(base_dir, conditions=None, exclude_malformed=False, exclude_empty=True):
    exps = experiment.get_all_experiments(base_dir, conditions)

    counts = {}
    no_outcomes = []

    for exp in exps:
        exp_counts = exp.load_outcome_counts()
        if exp_counts is None:
            no_outcomes.append((exp.group, exp.name))
        else:
            counts[exp.group, exp.name] = exp_counts

    if no_outcomes:
        no_outcomes_string = '\n'.join(f'\t{group}: {name}' for group, name in no_outcomes)
        print(f'Warning: can\'t find outcome counts for\n{no_outcomes_string}') 

    df = pd.DataFrame(counts).fillna(0)

    # Sort order for outcomes is defined in the relevant layout module.
    layout_modules = {exp.layout_module for exp in exps}
    
    if len(layout_modules) > 1:
        raise ValueError('Can\'t make table for experiments with inconsistent layout modules.')
    
    layout_module = layout_modules.pop()
    
    df = df.reindex(layout_module.full_index, fill_value=0)

    if exclude_malformed:
        df = df.drop('malformed layout', axis='index', level=0, errors='ignore')
        totals_row_label = totals_relevant_row_label
    else:
        totals_row_label = totals_all_row_label

    if exclude_empty:
        empty_rows = df.index[df.sum(axis=1) == 0].values
        df = df.drop(empty_rows, axis='index', errors='ignore')
    
    totals = df.sum(axis=0)
    totals_row = pd.DataFrame.from_dict({totals_row_label: totals}, orient='index')
    
    df = pd.concat([totals_row, df]).astype(int)
    df.index.names = (None, None)

    # Sort by group and sample name.
    df = df.sort_index(axis=1)

    return df

def calculate_performance_metrics(base_dir, conditions=None):
    full_counts = load_counts(base_dir, conditions=conditions)
    counts = full_counts.drop([totals_all_row_label, totals_relevant_row_label], axis='index', errors='ignore').sum(level=0)

    not_real_cell_categories = [
        'malformed layout',
    ]

    real_cells = counts.drop(not_real_cell_categories)

    all_edit_categories = [cat for cat in real_cells.index if cat != 'WT']

    all_integration_categories = [
        'HDR',
        'blunt misintegration',
        'complex misintegration',
        'concatenated misintegration',
        'incomplete HDR',
    ]

    # reindex to handle possibly missing keys
    HDR_counts = real_cells.reindex(['HDR'], fill_value=0).loc['HDR']
    edit_counts = real_cells.reindex(all_edit_categories, fill_value=0)
    integration_counts = real_cells.reindex(all_integration_categories, fill_value=0)

    performance_metrics = pd.DataFrame({
        'HDR_rate': HDR_counts / real_cells.sum(),
        'specificity_edits': HDR_counts / edit_counts.sum(),
        'specificity_integrations': HDR_counts / integration_counts.sum(),
    })

    return performance_metrics

def png_bytes_to_URI(png_bytes):
    encoded = base64.b64encode(png_bytes).decode('UTF-8')
    URI = f"'data:image/png;base64,{encoded}'"
    return URI

def fn_to_URI(fn):
    im = PIL.Image.open(fn)
    im.load()
    return Image_to_png_URI(im)

def Image_to_png_URI(im):
    with io.BytesIO() as buf:
        im.save(buf, format='png')
        png_bytes = buf.getvalue()
        
    URI = png_bytes_to_URI(png_bytes)
    
    return URI, im.width, im.height

def fig_to_png_URI(fig):
    with io.BytesIO() as buffer:
        fig.savefig(buffer, format='png', bbox_inches='tight')
        png_bytes = buffer.getvalue()
        im = PIL.Image.open(buffer)
        im.load()
       
    URI = png_bytes_to_URI(png_bytes)
    
    return URI, im.width, im.height

link_template = '''\
<a 
    data-toggle="popover" 
    data-trigger="hover"
    data-html="true"
    data-placement="auto"
    data-content="<img width={width} height={height} src={URI}>"
    onclick="$('#{modal_id} .modal-body').html('<iframe width=&quot;100%&quot; height=&quot;100%&quot; frameborder=&quot;0&quot; scrolling=&quot;no&quot; allowtransparency=&quot;true&quot; src=&quot;{iframe_URL}&quot;></iframe>'); $('#{modal_id}').appendTo('body').modal();"
    style="text-decoration:none; color:black"
>
    {text}
</a>
'''

link_without_modal_template = '''\
<a 
    data-toggle="popover" 
    data-trigger="hover"
    data-html="true"
    data-placement="auto"
    data-content="<img width={width} height={height} src={URI}>"
    style="text-decoration:none; color:black"
>
    {text}
</a>
'''
modal_template = '''\
<div class="modal" tabindex="-1" id="{modal_id}" role="dialog">
    <div class="modal-dialog" style="width:90%; margin:auto">
        <div class="modal-content">
            <div class="modal-body" style="height:10000px">
            </div>
        </div>
    </div>
</div>
'''

class ModalMaker(object):
    def __init__(self):
        self.current_number = 0

    def get_next_id(self):
        next_id = 'modal_{:06d}'.format(self.current_number)
        self.current_number += 1
        return next_id
        
    def make_length(self, exp, outcome=None, inline_images=True):
        modal_id = self.get_next_id()
        
        svg_text = svg.length_plot_with_popovers(exp, outcome=outcome, container_selector=f'#{modal_id}', inline_images=inline_images)
        modal_div = modal_template.format(modal_id=modal_id, contents=svg_text, title=exp.name)
        
        return modal_div, modal_id

    def make_outcome(self):
        modal_id = self.get_next_id()
        modal_div = modal_template.format(modal_id=modal_id)
        
        return modal_div, modal_id
        
def make_table(base_dir,
               conditions=None,
               inline_images=False,
               show_details=True,
              ):
    df = load_counts(base_dir, conditions=conditions, exclude_malformed=(not show_details))
    if show_details:
        totals_row_label = totals_all_row_label
    else:
        totals_row_label = totals_relevant_row_label
    totals_row_label_collapsed = totals_row_label[1]

    totals = df.loc[totals_row_label]

    df = df.T

    # Hack to give the html the information it needs to build links to diagram htmls
    df.index = pd.MultiIndex.from_tuples([(g, f'{g}/{n}') for g, n in df.index.values])
    
    if not show_details:
        level_0 = list(df.columns.levels[0])
        level_0[0] = totals_row_label[1]
        df.columns = df.columns.set_levels(level_0, level=0)

        df = df.sum(axis=1, level=0)

    exps = experiment.get_all_experiments(base_dir, conditions=conditions, as_dictionary=True)

    modal_maker = ModalMaker()

    def link_maker(val, outcome, exp_group, exp_name):
        if val == 0:
            html = ''
        else:
            exp = exps[exp_group, exp_name]
            
            fraction = val / totals[(exp_group, exp_name)]

            if outcome == totals_row_label or outcome == totals_row_label_collapsed:
                text = '{:,}'.format(val)

                hover_image_fn = exp.fns['lengths_figure']

                relative_path = hover_image_fn.relative_to(exp.base_dir / 'results')

                hover_URI = str(relative_path)
                if hover_image_fn.exists():
                    with PIL.Image.open(hover_image_fn) as im:
                        width, height = im.size
                        width = width * 0.75
                        height = height * 0.75
                else:
                    width, height = 100, 100

                link = link_without_modal_template.format(text=text,
                                                            URI=hover_URI,
                                                            width=width,
                                                            height=height,
                                                            )

                html = link

            else:
                text = '{:.2%}'.format(fraction)

                hover_image_fn = exp.outcome_fns(outcome)['first_example']
                click_html_fn = exp.outcome_fns(outcome)['diagrams_html']
                
                if inline_images:
                    hover_URI, width, height = fn_to_URI(hover_image_fn)
                else:
                    relative_path = hover_image_fn.relative_to(exp.base_dir / 'results')
                    hover_URI = str(relative_path)
                    if hover_image_fn.exists():
                        with PIL.Image.open(hover_image_fn) as im:
                            width, height = im.size
                            width = width * 0.75
                            height = height * 0.75
                    else:
                        width, height = 100, 100

                relative_path = click_html_fn.relative_to(exp.base_dir / 'results')

                modal_div, modal_id = modal_maker.make_outcome()

                link = link_template.format(text=text,
                                            modal_id=modal_id,
                                            iframe_URL=relative_path,
                                            URI=hover_URI,
                                            width=width,
                                            height=height,
                                            URL=str(relative_path),
                                           )
                html = link + modal_div

        return html
    
    def bind_link_maker(exp_group, exp_name):
        bound = {}
        for outcome in df:
            bound[outcome] = functools.partial(link_maker, outcome=outcome, exp_group=exp_group, exp_name=exp_name)

        return bound

    styled = df.style

    styles = [
        dict(selector="th", props=[("border", "1px solid black")]),
        dict(selector="tr:hover", props=[("background-color", "#cccccc")]),
    ]
    
    for exp_group, group_and_name in df.index:
        _, exp_name = group_and_name.split('/')
        sl = pd.IndexSlice[[(exp_group, group_and_name)], :]
        styled = styled.format(bind_link_maker(exp_group, exp_name), subset=sl)
    
    styled = styled.set_properties(**{'border': '1px solid black'})
    for exp_group, group_and_name in df.index:
        _, exp_name = group_and_name.split('/')
        exp = experiment.Experiment(base_dir, exp_group, exp_name)
        # Note: as of pandas 0.22, col needs to be in brackets here so that
        # apply is ultimately called on a df, not a series, to prevent
        # TypeError: _bar_left() got an unexpected keyword argument 'axis'
        styled = styled.bar(subset=pd.IndexSlice[[(exp_group, group_and_name)], :], axis=1, color=exp.color)
        
    styled.set_table_styles(styles)
    
    return styled

def generate_html(base_dir, fn, conditions=None, show_details=True):
    nb = nbf.new_notebook()

    cell_contents = f'''\
import knock_knock.table

conditions = {conditions}
knock_knock.table.make_table('{base_dir}', conditions, show_details={show_details})
'''
    
    nb['cells'] = [nbf.new_code_cell(cell_contents)]

    nb['metadata'] = {'title': str(fn.name)}

    exporter = nbconvert.HTMLExporter(exclude_input=True, exclude_output_prompt=True)
    template_path = Path(os.path.realpath(__file__)).parent / 'modal_template.tpl'
    exporter.template_file = str(template_path)

    ep = nbconvert.preprocessors.ExecutePreprocessor(timeout=600, kernel_name='python3')
    ep.preprocess(nb, {})

    body, resources = exporter.from_notebook_node(nb)
    with open(fn, 'w') as fh:
        fh.write(body)



import subprocess
import sys
import warnings
from urllib.parse import urlparse
from pathlib import Path

import pandas as pd
import numpy as np
import pysam
import yaml

import Bio.SeqIO
import Bio.SeqUtils
from Bio import BiopythonWarning
from Bio.SeqFeature import SeqFeature, FeatureLocation
from Bio.Seq import Seq
from Bio.SeqRecord import SeqRecord
from Bio.Alphabet import generic_dna

from hits import fastq, mapping_tools, sam, genomes, utilities, sw
from knock_knock import target_info

def design_amplicon_primers_from_csv(base_dir, genome='hg19'):
    base_dir = Path(base_dir)
    csv_fn = base_dir / 'targets' / 'sgRNAs.csv'

    index_locations = target_info.locate_supplemental_indices(base_dir)
    if genome not in index_locations:
        print(f'Error: can\'t locate indices for {genome}')
        sys.exit(0)

    df = pd.read_csv(csv_fn).replace({np.nan: None})

    amplicon_primer_info = {}

    for _, row in df.iterrows():
        name = row['name']
        print(f'Designing {name}...')
        best_candidate = design_amplicon_primers(base_dir, row, genome)
        if best_candidate is not None:
            amplicon_primer_info[name] = {
                'flanking_sequence': best_candidate['target_seq'],
                'genome': genome,
                'ref_name': best_candidate['ref_name'],
                'min_cut_after': best_candidate['min_cut_after'],
                'max_cut_after': best_candidate['max_cut_after'],
            }

    amplicon_primer_info = pd.DataFrame(amplicon_primer_info).T
    amplicon_primer_info.index.name = 'name'

    column_order = ['flanking_sequence', 'genome', 'ref_name', 'min_cut_after', 'max_cut_after']
    amplicon_primer_info = amplicon_primer_info[column_order]

    final_csv_fn = base_dir / 'targets' / 'sgRNAs_flanking_sequence.csv'
    amplicon_primer_info.to_csv(final_csv_fn)

def design_amplicon_primers(base_dir, info, genome):
    base_dir = Path(base_dir)

    name = info['name']

    target_dir = base_dir / 'targets' / name
    target_dir.mkdir(parents=True, exist_ok=True)

    protospacer, *other_protospacers = info['sgRNA_sequence'].upper().split(';')
    
    protospacer_dir = target_dir / 'protospacer_alignment'
    protospacer_dir.mkdir(exist_ok=True)
    fastq_fn = protospacer_dir / 'protospacer.fastq'
    STAR_prefix = protospacer_dir / 'protospacer_'
    bam_fn = protospacer_dir / 'protospacer.bam'

    index_locations = target_info.locate_supplemental_indices(base_dir)
    STAR_index = index_locations[genome]['STAR']

    # Make a fastq file with a single read containing the protospacer sequence.
    
    with fastq_fn.open('w') as fh:
        quals = fastq.encode_sanger([40]*len(protospacer))
        read = fastq.Read('protospacer', protospacer, quals)
        fh.write(str(read))
        
    # Align the protospacer to the reference genome.
    mapping_tools.map_STAR(fastq_fn, STAR_index, STAR_prefix, mode='guide_alignment', bam_fn=bam_fn, sort=False)

    with pysam.AlignmentFile(bam_fn) as bam_fh:
        perfect_als = [al for al in bam_fh if not al.is_unmapped and sam.total_edit_distance(al) == 0]
    
    region_fetcher = genomes.build_region_fetcher(index_locations[genome]['fasta'])

    def evaluate_candidate(al):
        results = {
            'location': f'{al.reference_name} {al.reference_start:,} {sam.get_strand(al)}',
            'ref_name': al.reference_name,
            'cut_afters': [],
        }

        full_window_around = 5000

        full_around = region_fetcher(al.reference_name, al.reference_start - full_window_around, al.reference_end + full_window_around).upper()

        if sam.get_strand(al) == '+':
            ps_seq = protospacer
            ps_strand = 1
        else:
            ps_seq = utilities.reverse_complement(protospacer)
            ps_strand  = -1
        
        ps_start = full_around.index(ps_seq)

        protospacer_locations = [(ps_seq, ps_start, ps_strand)]

        for other_protospacer in other_protospacers:
            if other_protospacer in full_around:
                ps_seq = other_protospacer
                ps_strand = 1
            else:
                ps_seq =  utilities.reverse_complement(other_protospacer)
                if ps_seq not in full_around:
                    results['failed'] = f'protospacer {other_protospacer} not present near protospacer {protospacer}'
                    return results
                ps_strand = -1

            ps_start = full_around.index(ps_seq)
            protospacer_locations.append((ps_seq, ps_start, ps_strand))

        for ps_seq, ps_start, ps_strand in protospacer_locations:
            if ps_strand == 1:
                PAM_offset = len(protospacer)
                PAM_transform = utilities.identity
                cut_after = al.reference_start - full_window_around + ps_start + PAM_offset - 3
            else:
                PAM_offset = -3
                PAM_transform = utilities.reverse_complement
                cut_after = al.reference_start - full_window_around + ps_start + 2

            results['cut_afters'].append(cut_after)

            PAM_start = ps_start + PAM_offset
            PAM = PAM_transform(full_around[PAM_start:PAM_start + 3])
            pattern, *matches = Bio.SeqUtils.nt_search(PAM, 'NGG')

            if 0 not in matches:
                # Note: this could incorrectly fail if there are multiple exact matches for an other_protospacer
                # in full_around.
                results['failed'] = f'bad PAM: {PAM} next to {ps_seq} (strand {ps_strand})'
                return results

        min_start = min(ps_start for ps_seq, ps_start, ps_strand in protospacer_locations)
        max_start = max(ps_start for ps_seq, ps_start, ps_strand in protospacer_locations)

        results['min_cut_after'] = min(results['cut_afters'])
        results['max_cut_after'] = max(results['cut_afters'])
        
        final_window_around = 500    

        final_start = min_start - final_window_around
        final_end = max_start + final_window_around

        target_seq = full_around[final_start:final_end]
        results['target_seq'] = target_seq

        return results

    good_candidates = []
    bad_candidates = []
    
    for al in perfect_als:
        results = evaluate_candidate(al)
        if 'failed' in results:
            bad_candidates.append(results)
        else:
            good_candidates.append(results)

    if len(good_candidates) == 0:
        if len(bad_candidates) == 0:
            print(f'Error building {name}: no perfect matches to sgRNA {protospacer} found in {genome}')
            return 

        else:
            print(f'Error building {name}: no valid genomic locations for {name}')

            for results in bad_candidates:
                print(f'\t{results["location"]}: {results["failed"]}')

            return 

    elif len(good_candidates) > 1:
        print(f'Warning: multiple valid genomic locations for {name}:')
        for results in good_candidates:
            print(f'\t{results["location"]}')
        best_candidate = good_candidates[0]
        print(f'Arbitrarily choosing {best_candidate["location"]}')
    else:
        best_candidate = good_candidates[0]

    return best_candidate

def identify_homology_arms(donor_seq, target_seq, cut_after):
    required_match_length = 15
    
    header = pysam.AlignmentHeader.from_references(['donor', 'target'], [len(donor_seq), len(target_seq)])
    mapper = sw.SeedAndExtender(donor_seq.encode(), 8, header, 'donor')
    
    target_bytes = target_seq.encode()
    
    alignments = {
        'before_cut': [],
        'after_cut': [],
    }

    seed_starts = {
        'before_cut': range(cut_after - required_match_length, 0, -1),
        'after_cut': range(cut_after, len(target_seq) - required_match_length),
    }

    for side in ['before_cut', 'after_cut']:
        for seed_start in seed_starts[side]:  
            alignments[side] = mapper.seed_and_extend(target_bytes, seed_start, seed_start + required_match_length, 'target')
            if alignments[side]:
                break

        else:
            results = {'failed': f'cannot locate homology arm on {side}'}
            return results
        
    possible_HA_boundaries = []
    
    for before_al in alignments['before_cut']:
        for after_al in alignments['after_cut']:
            if sam.get_strand(before_al) == sam.get_strand(after_al):
                strand = sam.get_strand(before_al)
                if strand == '+':
                    if before_al.reference_end < after_al.reference_start:
                        possible_HA_boundaries.append((donor_seq, before_al.reference_start, after_al.reference_end))
                elif strand == '-':
                    if before_al.reference_start > after_al.reference_end:
                        flipped = utilities.reverse_complement(donor_seq)
                        start = len(donor_seq) - 1 - (before_al.reference_end - 1)
                        end = len(donor_seq) - 1 - after_al.reference_start + 1
                        possible_HA_boundaries.append((flipped, start, end))

         
    possible_HAs = []
    for possibly_flipped_donor_seq, HA_start, HA_end in possible_HA_boundaries:
        donor_window = possibly_flipped_donor_seq[HA_start:HA_end]

        donor_prefix = donor_window[:required_match_length]

        if donor_prefix not in target_seq:
            raise ValueError

        donor_suffix = donor_window[-required_match_length:]

        if donor_suffix not in target_seq:
            raise ValueError

        target_HA_start = target_seq.index(donor_prefix)
        target_HA_end = target_seq.index(donor_suffix) + len(donor_suffix)

        relevant_target_seq = target_seq[target_HA_start:target_HA_end]

        total_HA_length = target_HA_end - target_HA_start

        mismatches_before_deletion = np.cumsum([t != d for t, d in zip(relevant_target_seq, donor_window)])

        flipped_target = relevant_target_seq[::-1]
        flipped_donor = donor_window[::-1]
        mismatches_after_deletion = np.cumsum([0] + [t != d for t, d in zip(flipped_target, flipped_donor)][:-1])[::-1]

        total_mismatches = mismatches_before_deletion + mismatches_after_deletion

        last_index_in_HA_1 = int(np.argmin(total_mismatches))
        min_mismatches = total_mismatches[last_index_in_HA_1]

        lengths = {}
        lengths['HA_1'] = last_index_in_HA_1 + 1
        lengths['HA_2'] = total_HA_length - lengths['HA_1']
        lengths['donor_specific'] = len(donor_seq) - total_HA_length
        
        info = {
            'possibly_flipped_donor_seq': possibly_flipped_donor_seq,
            'donor_HA_start': HA_start,
            'donor_HA_end': HA_end,
            'target_HA_start': target_HA_start,
            'target_HA_end': target_HA_end,
            'lengths': lengths,
        }
        possible_HAs.append((min_mismatches, info))
        
    if not possible_HAs:
        results = {'failed': 'cannot locate homology arms'}
    else:
        results = min(possible_HAs)[1]

    return results

def build_target_info(base_dir, info, all_index_locations):
    ''' info should have keys:
            sgRNA_sequence
            amplicon_primers
        optional keys:
            donor_sequence
            nonhomologous_donor_sequence
            extra_sequences
    '''
    genome = info['genome']
    if info['genome'] not in all_index_locations:
        print(f'Error: can\'t locate indices for {genome}')
        sys.exit(0)
    else:
        index_locations = all_index_locations[genome]

    base_dir = Path(base_dir)

    name = info['name']

    donor_info = info.get('donor_sequence')
    if donor_info is None:
        donor_name = None
        donor_seq = None
    else:
        donor_name, donor_seq = donor_info
        if donor_name is None:
            donor_name = f'{name}_donor'

    if donor_seq is None:
        has_donor = False
    else:
        has_donor = True

    if info['donor_type'] is None:
        donor_type = None
    else:
        _, donor_type = info['donor_type']

    nh_donor_info = info.get('nonhomologous_donor_sequence')
    if nh_donor_info is None:
        nh_donor_name = None
        nh_donor_seq = None
    else:
        nh_donor_name, nh_donor_seq = nh_donor_info
        if nh_donor_name is None:
            nh_donor_name = f'{name}_NH_donor'

    if nh_donor_seq is None:
        has_nh_donor = False
    else:
        has_nh_donor = True

    target_dir = base_dir / 'targets' / name
    target_dir.mkdir(parents=True, exist_ok=True)
    
    protospacer, *other_protospacers = info['sgRNA_sequence']
    primers_name, primers = info['amplicon_primers']
    primers = primers.split(';')

    if primers_name is None:
        target_name = name
    else:
        target_name = primers_name

    protospacer_dir = target_dir / 'protospacer_alignment'
    protospacer_dir.mkdir(exist_ok=True)
    fastq_fn = protospacer_dir / 'protospacer.fastq'
    STAR_prefix = protospacer_dir / 'protospacer_'
    bam_fn = protospacer_dir / 'protospacer.bam'

    STAR_index = index_locations['STAR']

    gb_fns = {
        'target': target_dir / f'{target_name}.gb',
        'donor': target_dir / f'{donor_name}.gb',
        'nh_donor': target_dir / f'{nh_donor_name}.gb',
    }

    # Make a fastq file with a single read containing the protospacer sequence.
    protospacer_name, protospacer_seq = protospacer
    
    with fastq_fn.open('w') as fh:
        quals = fastq.encode_sanger([40]*len(protospacer_seq))
        read = fastq.Read('protospacer', protospacer_seq, quals)
        fh.write(str(read))
        
    # Align the protospacer to the reference genome.
    mapping_tools.map_STAR(fastq_fn, STAR_index, STAR_prefix, mode='guide_alignment', bam_fn=bam_fn, sort=False)

    with pysam.AlignmentFile(bam_fn) as bam_fh:
        perfect_als = [al for al in bam_fh if not al.is_unmapped and sam.total_edit_distance(al) == 0]
    
    region_fetcher = genomes.build_region_fetcher(index_locations['fasta'])
    
    def evaluate_candidate(al):
        results = {
            'location': f'{al.reference_name} {al.reference_start:,} {sam.get_strand(al)}',
        }

        full_window_around = 5000

        full_around = region_fetcher(al.reference_name, al.reference_start - full_window_around, al.reference_end + full_window_around).upper()

        if sam.get_strand(al) == '+':
            ps_seq = protospacer_seq
            ps_strand = 1
        else:
            ps_seq = utilities.reverse_complement(protospacer_seq)
            ps_strand  = -1

        ps_start = full_around.index(ps_seq)

        protospacer_locations = [(ps_seq, ps_start, ps_strand)]

        for other_protospacer_name, other_protospacer_seq in other_protospacers:
            if other_protospacer_seq in full_around:
                ps_seq = other_protospacer_seq
                ps_strand = 1
            else:
                ps_seq =  utilities.reverse_complement(other_protospacer_seq)
                if ps_seq not in full_around:
                    results['failed'] = f'protospacer {other_protospacer_seq} not present near protospacer {protospacer_seq}'
                    return results
                ps_strand = -1

            ps_start = full_around.index(ps_seq)
            protospacer_locations.append((ps_seq, ps_start, ps_strand))

        for ps_seq, ps_start, ps_strand in protospacer_locations:
            PAM_pattern = 'NGG'

            if ps_strand == 1:
                PAM_offset = len(ps_seq)
                PAM_transform = utilities.identity
            else:
                PAM_offset = -len(PAM_pattern)
                PAM_transform = utilities.reverse_complement

            PAM_start = ps_start + PAM_offset
            PAM = PAM_transform(full_around[PAM_start:PAM_start + len(PAM_pattern)])
            pattern, *matches = Bio.SeqUtils.nt_search(PAM, PAM_pattern)

            if 0 not in matches:
                # Note: this could incorrectly fail if there are multiple exact matches for an other_protospacer
                # in full_around.
                results['failed'] = f'bad PAM: {PAM} next to {ps_seq} (strand {ps_strand})'
                return results

        if primers[0] in full_around:
            final_fwd_primer = primers[0]
            final_rev_primer = utilities.reverse_complement(primers[1])
            if final_rev_primer not in full_around:
                results['failed'] = f'primer {primers[1]} not present near protospacer'
                return results
        else:
            final_fwd_primer = primers[1]
            final_rev_primer = utilities.reverse_complement(primers[0])

            if final_fwd_primer not in full_around:
                results['failed'] = f'primer {primers[1]} not present near protospacer'
                return results

            if final_rev_primer not in full_around:
                results['failed'] = f'primer {primers[0]} not present near protospacer'
                return results

        fwd_start = full_around.index(final_fwd_primer)
        rev_start = full_around.index(final_rev_primer)
        
        if fwd_start >= rev_start:
            results['failed'] = f'primer don\'t flank protospacer'
            return results
        
        final_window_around = 500    

        offset = fwd_start - final_window_around

        final_start = fwd_start - final_window_around
        final_end = rev_start + len(final_rev_primer) + final_window_around

        target_seq = full_around[final_start:final_end]

        colors = {
            'HA_1': '#c7b0e3',
            'HA_2': '#85dae9',
            'forward_primer': '#ff9ccd',
            'reverse_primer': '#9eafd2',
            'sgRNA': '#c6c9d1',
            'donor_specific': '#b1ff67',
            'PCR_adapter_1': '#F8D3A9',
            'PCR_adapter_2': '#D59687',
        }
        
        target_features = [
            SeqFeature(location=FeatureLocation(fwd_start - offset, fwd_start - offset + len(final_fwd_primer), strand=1),
                       id='forward_primer',
                       type='misc_feature',
                       qualifiers={'label': 'forward_primer',
                                   'ApEinfo_fwdcolor': colors['forward_primer'],
                                  },
                      ),
            SeqFeature(location=FeatureLocation(rev_start - offset, rev_start - offset + len(final_rev_primer), strand=-1),
                       id='reverse_primer',
                       type='misc_feature',
                       qualifiers={'label': 'reverse_primer',
                                   'ApEinfo_fwdcolor': colors['reverse_primer'],
                                  },
                      ),
        ]

        sgRNA_features = []
        for sgRNA_i, (ps_seq, ps_start, ps_strand) in enumerate(protospacer_locations):
            sgRNA_feature = SeqFeature(location=FeatureLocation(ps_start - offset, ps_start - offset + len(ps_seq), strand=ps_strand),
                                       id=f'sgRNA_{sgRNA_i}',
                                       type='sgRNA_SpCas9',
                                       qualifiers={'label': f'sgRNA_{sgRNA_i}',
                                                   'ApEinfo_fwdcolor': colors['sgRNA'],
                                                   },
                                       )
            target_features.append(sgRNA_feature)
            sgRNA_features.append(sgRNA_feature)

        results['gb_Records'] = {}

        if has_donor:
            # Identify the homology arms.

            if len(sgRNA_features) > 1:
                results['failed'] = 'multiple sgRNAs and a donor are not supported'
            else:
                sgRNA_feature = sgRNA_features[0]
                if sgRNA_feature.strand == 1:
                    # sgRNA_feature.end is the first nt of the PAM
                    cut_after = sgRNA_feature.location.end + target_info.effectors['SpCas9'].cut_after_offset
                else:
                    # sgRNA_feature.start - 1 is the first nt of the PAM
                    cut_after = sgRNA_feature.location.start - 1 - target_info.effectors['SpCas9'].cut_after_offset - 1

            HA_info = identify_homology_arms(donor_seq, target_seq, cut_after)
            if 'failed' in HA_info:
                results['failed'] = HA_info['failed']
                return results

            lengths = HA_info['lengths']

            starts = {
                'HA_1': HA_info['donor_HA_start'],
                'donor_specific': HA_info['donor_HA_start'] + lengths['HA_1'],
                'HA_2': HA_info['donor_HA_end'] - lengths['HA_2'],
            }
            ends = {
                'HA_1': starts['HA_1'] + lengths['HA_1'],
                'donor_specific': starts['HA_2'],
                'HA_2': starts['HA_2'] + lengths['HA_2'],
            }

            if donor_type == 'PCR':
                if starts['HA_1'] != 0:
                    starts['PCR_adapter_1'] = 0
                    ends['PCR_adapter_1'] = starts['HA_1']

                if ends['HA_2'] != len(donor_seq):
                    starts['PCR_adapter_2'] = ends['HA_2']
                    ends['PCR_adapter_2'] = len(donor_seq)
        
            donor_features = [
                SeqFeature(location=FeatureLocation(starts[key], ends[key], strand=1),
                           id=key,
                           type='misc_feature',
                           qualifiers={'label': key,
                                       'ApEinfo_fwdcolor': colors[key],
                                      },
                        )
                for key in starts
            ]

            donor_Seq = Seq(HA_info['possibly_flipped_donor_seq'], generic_dna)
            donor_Record = SeqRecord(donor_Seq, name=donor_name, features=donor_features)
            results['gb_Records']['donor'] = donor_Record

            target_features.extend([
                SeqFeature(location=FeatureLocation(HA_info['target_HA_start'], HA_info['target_HA_start'] + lengths['HA_1'], strand=1),
                        id='HA_1',
                        type='misc_feature',
                        qualifiers={'label': 'HA_1',
                                    'ApEinfo_fwdcolor': colors['HA_1'],
                                    },
                        ),
                SeqFeature(location=FeatureLocation(HA_info['target_HA_end'] - lengths['HA_2'], HA_info['target_HA_end'], strand=1),
                        id='HA_2',
                        type='misc_feature',
                        qualifiers={'label': 'HA_2',
                                    'ApEinfo_fwdcolor': colors['HA_2'],
                                    },
                        ),
            ])

        target_Seq = Seq(target_seq, generic_dna)
        target_Record = SeqRecord(target_Seq, name=target_name, features=target_features)
        results['gb_Records']['target'] = target_Record
        
        if has_nh_donor:
            nh_donor_Seq = Seq(nh_donor_seq, generic_dna)
            nh_donor_Record = SeqRecord(nh_donor_Seq, name=nh_donor_name)
            results['gb_Records']['nh_donor'] = nh_donor_Record

        return results
    
    good_candidates = []
    bad_candidates = []
    
    for al in perfect_als:
        results = evaluate_candidate(al)
        if 'failed' in results:
            bad_candidates.append(results)
        else:
            good_candidates.append(results)
    
    if len(good_candidates) == 0:
        if len(bad_candidates) == 0:
            print(f'Error building {name}: no perfect matches to sgRNA {protospacer} found in {genome}')
            return 

        else:
            print(f'Error building {name}: no valid genomic locations for {name}')

            for results in bad_candidates:
                print(f'\t{results["location"]}: {results["failed"]}')

            return 

    elif len(good_candidates) > 1:
        print(f'Warning: multiple valid genomic locations for {name}:')
        for results in good_candidates:
            print(f'\t{results["location"]}')
        best_candidate = good_candidates[0]
        print(f'Arbitrarily choosing {best_candidate["location"]}')
    else:
        best_candidate = good_candidates[0]

    truncated_name_i = 0
    with warnings.catch_warnings():
        warnings.filterwarnings('ignore', category=BiopythonWarning)

        for which_seq, Record in best_candidate['gb_Records'].items(): 
            try:
                Bio.SeqIO.write(Record, gb_fns[which_seq], 'genbank')
            except ValueError:
                # locus line too long, can't write genbank file with BioPython
                old_name = Record.name

                truncated_name = f'{Record.name[:11]}_{truncated_name_i}'
                Record.name = truncated_name
                Bio.SeqIO.write(Record, gb_fns[which_seq], 'genbank')

                Record.name = old_name

                truncated_name_i += 1

    manifest_fn = target_dir / 'manifest.yaml'

    sources = [target_name]
    if has_donor:
        sources.append(donor_name)

    extra_Records = []
    if info.get('extra_sequences') is not None:
        for extra_seq_name, extra_seq in info['extra_sequences']:
            sources.append(extra_seq_name)

            extra_Records.append(SeqRecord(extra_seq, name=extra_seq_name))
        
    manifest = {
        'sources': sources,
        'target': target_name,
    }
    if has_donor:
        manifest['donor'] = donor_name
        manifest['donor_specific'] = 'donor_specific'
        if donor_type is not None:
            manifest['donor_type'] = donor_type

    if has_nh_donor:
        manifest['nonhomologous_donor'] = nh_donor_name

    manifest_fn.write_text(yaml.dump(manifest, default_flow_style=False))
        
    gb_records = list(best_candidate['gb_Records'].values()) + extra_Records
    ti = target_info.TargetInfo(base_dir, name, gb_records=gb_records)
    ti.make_references()
    ti.identify_degenerate_indels()

def build_target_infos_from_csv(base_dir):
    base_dir = Path(base_dir)
    csv_fn = base_dir / 'targets' / 'targets.csv'

    indices = target_info.locate_supplemental_indices(base_dir)

    targets_df = pd.read_csv(csv_fn, comment='#', index_col='name').replace({np.nan: None})

    # Fill in values for sequences that are specified by name.

    registry = {}

    sgRNA_fn = base_dir / 'targets' / 'sgRNAs.csv'

    if sgRNA_fn.exists():
        registry['sgRNA_sequence'] = pd.read_csv(sgRNA_fn, index_col='name', squeeze=True)
    else:
        registry['sgRNA_sequence'] = {}

    amplicon_primers_fn = base_dir / 'targets' / 'amplicon_primers.csv'

    if amplicon_primers_fn.exists():
        registry['amplicon_primers'] = pd.read_csv(amplicon_primers_fn, index_col='name', squeeze=True)
    else:
        registry['amplicon_primers'] = {}

    extra_sequences_fn = base_dir / 'targets' / 'extra_sequences.csv'
    if extra_sequences_fn.exists():
        registry['extra_sequence'] = pd.read_csv(extra_sequences_fn, index_col='name', squeeze=True)
    else:
        registry['extra_sequence'] = {}

    donors_fn = base_dir / 'targets' / 'donors.csv'

    if donors_fn.exists():
        donors = pd.read_csv(donors_fn, index_col='name')
        registry['donor_sequence'] = donors['donor_sequence']
        registry['donor_type'] = donors['donor_type']
    else:
        registry['donor_sequence'] = {}
        registry['donor_type'] = {}

    def lookup(row, column_to_lookup, registry_column, validate_sequence=True, multiple_lookups=False):
        value_to_lookup = row.get(column_to_lookup)
        if value_to_lookup is None:
            return None

        if multiple_lookups:
            values_to_lookup = value_to_lookup.split(';')
        else:
            values_to_lookup = [value_to_lookup]

        registered_values = registry[registry_column]
        valid_chars = set('TCAGN;')

        looked_up = []
        for value_to_lookup in values_to_lookup:
            if value_to_lookup in registered_values:
                value_name = value_to_lookup
                seq = registered_values[value_to_lookup]
                possible_error_message = f'invalid char in {row.name} {column_to_lookup} registry entry {value_to_lookup}\n{seq}'
            else:
                value_name = None
                seq = value_to_lookup
                possible_error_message = f'invalid char in {row.name}: {seq} \n Registered names: {registered_values}'

            if seq is not None and validate_sequence:
                seq = seq.upper()
                invalid_chars = set(seq) - valid_chars
                if invalid_chars:
                    print(possible_error_message)
                    print(invalid_chars)
                    sys.exit(0)

            looked_up.append((value_name, seq))

        if not multiple_lookups:
            looked_up = looked_up[0]
            
        return looked_up

    for target_name, row in targets_df.iterrows():
        info = {
            'name': target_name,
            'donor_sequence': lookup(row, 'donor_sequence', 'donor_sequence'),
            'sgRNA_sequence': lookup(row, 'sgRNA_sequence', 'sgRNA_sequence', multiple_lookups=True),
            'extra_sequences': lookup(row, 'extra_sequences', 'extra_sequence', multiple_lookups=True),
            'amplicon_primers': lookup(row, 'amplicon_primers', 'amplicon_primers'),
            'nonhomologous_donor_sequence': lookup(row, 'nonhomologous_donor_sequence', 'donor_sequence'),
            'donor_type': lookup(row, 'donor_sequence', 'donor_type', validate_sequence=False),
            'genome': row['genome'],
        }

        print(f'Building {target_name}...')
        build_target_info(base_dir, info, indices)

def build_indices(base_dir, name, num_threads=1):
    base_dir = Path(base_dir)

    print(f'Building indices for {name}')
    fasta_dir = base_dir / 'indices' / name / 'fasta'

    fasta_fns = genomes.get_all_fasta_file_names(fasta_dir)
    if len(fasta_fns) == 0:
        raise ValueError(f'No fasta files found in {fasta_dir}')
    elif len(fasta_fns) > 1:
        raise ValueError(f'Can only build minimap2 index from a single fasta file')

    print('Indexing fastas...')
    genomes.make_fais(fasta_dir)

    fasta_fn = fasta_fns[0]

    print('Building STAR index...')
    STAR_dir = base_dir / 'indices' / name / 'STAR'
    STAR_dir.mkdir(exist_ok=True)
    mapping_tools.build_STAR_index([fasta_fn], STAR_dir, num_threads=num_threads, RAM_limit=int(4e10))

    print('Building minimap2 index...')
    minimap2_dir = base_dir / 'indices' / name / 'minimap2'
    minimap2_dir.mkdir(exist_ok=True)
    minimap2_index_fn = minimap2_dir / f'{name}.mmi'
    mapping_tools.build_minimap2_index(fasta_fn, minimap2_index_fn)

def download_genome_and_build_indices(base_dir, genome_name, num_threads=8):
    urls = {
        'hg38': 'http://hgdownload.cse.ucsc.edu/goldenPath/hg38/bigZips/hg38.fa.gz',
        'mm10': 'ftp://ftp.ensembl.org/pub/release-98/fasta/mus_musculus/dna/Mus_musculus.GRCm38.dna.toplevel.fa.gz',
        'e_coli': 'ftp://ftp.ensemblgenomes.org/pub/bacteria/release-44/fasta/bacteria_0_collection/escherichia_coli_str_k_12_substr_mg1655/dna/Escherichia_coli_str_k_12_substr_mg1655.ASM584v2.dna.chromosome.Chromosome.fa.gz',
    }

    if genome_name not in urls:
        print(f'No URL known for {genome_name}. Options are:')
        for gn in sorted(urls):
            print(f'\t- {gn}')
        sys.exit(0)

    base_dir = Path(base_dir)
    genome_dir = base_dir / 'indices' / genome_name
    fasta_dir = genome_dir / 'fasta'

    print(f'Downloading {genome_name}...')

    wget_command = [
        'wget',
        '--quiet',
        urls[genome_name],
        '-P', str(fasta_dir),
    ]
    subprocess.run(wget_command, check=True)

    print('Uncompressing...')

    file_name = Path(urlparse(urls[genome_name]).path).name

    gunzip_command = [
        'gunzip',  str(fasta_dir / file_name),
    ]
    subprocess.run(gunzip_command, check=True)

    build_indices(base_dir, genome_name, num_threads=num_threads)



import bokeh
import bokeh.plotting
import bokeh.io
import numpy as np

from . import experiment

bokeh.io.output_notebook()

def range_picker(base_dir, inital_dataset=None):
    exps = {exp.name: exp for exp in experiment.get_all_experiments(base_dir)}
    if inital_dataset is None:
        inital_dataset = list(exps)[0]

    name_to_color = {name: exps[name].color for name in exps}

    all_ys = {name: exps[name].read_lengths for name in exps}
    max_length = max(len(ys) for ys in all_ys.values())
    max_y = max(max(ys) for ys in all_ys.values()) * 1.02
    xs = np.arange(max_length)
    for name in all_ys:
        padded = np.zeros(max_length, dtype=int)
        ys = all_ys[name]
        padded[:len(ys)] = ys
        all_ys[name] = padded

    f = bokeh.plotting.Figure(
        plot_width=1500,
        plot_height=600,
        lod_threshold=20000,
        active_scroll='wheel_zoom',
    )
    
    source = bokeh.models.ColumnDataSource(data={
        'x': xs,
        'y': all_ys[inital_dataset],
        **all_ys,
    })
    
    line = f.line(x='x', y='y', source=source, color=name_to_color[inital_dataset])

    f.x_range = bokeh.models.Range1d(start=0, end=max_length)
    f.y_range = bokeh.models.Range1d(start=0, end=max(all_ys[inital_dataset] * 1.02))

    bound_at_zero = bokeh.models.CustomJS.from_coffeescript('cb_obj.start = 0 if cb_obj.start < 0')
    fix_at_zero = bokeh.models.CustomJS.from_coffeescript('cb_obj.start = 0 if cb_obj.start != 0')
    y_bound_at_max = bokeh.models.CustomJS.from_coffeescript('cb_obj.end = {max} if cb_obj.end > {max}'.format(max=max_y))
    x_bound_at_max = bokeh.models.CustomJS.from_coffeescript('cb_obj.end = {max} if cb_obj.end > {max}'.format(max=max_length))

    f.x_range.js_on_change('start', bound_at_zero)
    f.x_range.js_on_change('end', x_bound_at_max)
    f.y_range.js_on_change('start', fix_at_zero)
    f.y_range.js_on_change('end', y_bound_at_max)

    q_data_source = bokeh.models.ColumnDataSource(data=dict(
        top=[],
        bottom=[],
        left=[],
        right=[],
    ))

    q = f.quad(
        top='top',
        bottom='bottom',
        left='left',
        right='right',
        source=q_data_source,
        color='black',
        alpha=0.15,
    )

    save_code = '''\
    data = data_source.data
    # learned from http://stackoverflow.com/questions/14964035/how-to-export-javascript-array-info-to-csv-on-client-side

    length = data.left.length
    lines = (data.left[i] + '\t' + data.right[i] for i in [0...length])

    csv_content = "data:text/csv;charset=utf-8," + lines.join('\\n')
    encoded = encodeURI(csv_content)

    link = document.createElement('a')
    link.setAttribute('href', encoded)
    link.setAttribute('download', 'manual_length_ranges.csv')
    link.click()
    '''
    save_button = bokeh.models.Button(label='save x ranges')
    save_button.js_on_click(bokeh.models.CustomJS.from_coffeescript(save_code, args={'data_source': q.data_source}))

    t = bokeh.models.BoxSelectTool(callback=bokeh.models.CustomJS.from_coffeescript('''
        data = data_source.data
        console.log data
        data['bottom'].push 0
        data['top'].push {max_y}
        data['left'].push Math.round(cb_data.geometry.x0)
        data['right'].push Math.round(cb_data.geometry.x1)
        data_source.change.emit()
    '''.format(max_y=max_y), args={'data_source': q.data_source}))

    f.add_tools(t)
    f.grid.visible = False

    select = bokeh.models.Select(title='Dataset:',
                                 value=inital_dataset,
                                 options=sorted(exps),
                                )
    
    select.js_on_change('value', bokeh.models.CustomJS.from_coffeescript(
        '''\
        data = data_source.data
        name = cb_obj.value
        data['y'] = data[name]

        name_to_color = {name_to_color}
        line.glyph.line_color = name_to_color[name]

        y_range.end = Math.max(...data['y']) * 1.02

        data_source.change.emit()

        q_data_source.data[key] = [] for key in ['top', 'bottom', 'left', 'right']
        q_data_source.change.emit()
        '''.format(name_to_color=name_to_color),
        args=dict(data_source=source, line=line, y_range=f.y_range, q_data_source=q_data_source)))

    bokeh.io.show(bokeh.layouts.row(children=[f, bokeh.layouts.column(children=[select, save_button])]))



import re
import html
import io
import xml.etree.ElementTree as ET

import PIL
import matplotlib
matplotlib.use('Agg', warn=False)
import matplotlib.pyplot as plt

from . import table, layout

before_svg_template = '''\
<head>
<title>{title}</title>
<link rel="stylesheet" href="https://stackpath.bootstrapcdn.com/bootstrap/4.3.1/css/bootstrap.min.css" integrity="sha384-ggOyR0iXCbMQv3Xipma34MD+dH/1fQ784/j6cY/iJTQUOhcWr7x9JvoRxT2MZw1T" crossorigin="anonymous">

<style>
.popover {{
    max-width: 100%;
}}
</style>

</head>
<body>
<div style="height:5000px;text-align:center">\
'''.format

after_svg = '''\
</div>
<script src="https://code.jquery.com/jquery-3.3.1.slim.min.js" integrity="sha384-q8i/X+965DzO0rT7abK41JStQIAqVgRVzpbzo5smXKp4YfRvH+8abtTE1Pi6jizo" crossorigin="anonymous"></script>
<script src="https://cdnjs.cloudflare.com/ajax/libs/popper.js/1.14.7/umd/popper.min.js" integrity="sha384-UO2eT0CpHqdSJQ6hJty5KVphtPhzWj9WO1clHTMGa3JDZwrnQq4sF86dIHNDz0W1" crossorigin="anonymous"></script>
<script src="https://stackpath.bootstrapcdn.com/bootstrap/4.3.1/js/bootstrap.min.js" integrity="sha384-JjSmVgyd0p3pXB1rRibZUAYoIIy6OrQ6VrjIEaFf/nJGzIxFDsf4x0xIM+B07jRM" crossorigin="anonymous"></script>

<script>
    $(document).ready(function(){
        $('[id^=length_range]').hide();

        $('[data-toggle="popover"]').popover();

        $('[data-toggle="popover"]').on('show.bs.popover', function() {
          $("path", this).attr('stroke-opacity', '0.5');
        });

        $('[data-toggle="popover"]').on('hide.bs.popover', function() {
          $("path", this).attr('stroke-opacity', '0.0');
        });
    });

    document.onkeydown = function(evt) {
        evt = evt || window.event;
        if (evt.keyCode == 27) {
            $('[data-toggle="popover"').popover('hide')
            $('[id^=length_range]').hide();

            $('[id^=outcome_category]').find('path').css('stroke-opacity', '1');

            $('[id^=line_highlighted]').find('path').css('stroke-width', '1');
            $('[id^=line_nonhighlighted]').find('path').css('stroke-width', '1');

            $('[id^=line_highlighted]').find('path').css('stroke-opacity', '1');
            $('[id^=line_nonhighlighted_6]').find('path').css('stroke-opacity', '0.6');
            $('[id^=line_nonhighlighted_3]').find('path').css('stroke-opacity', '0.3');
            $('[id^=line_nonhighlighted_05]').find('path').css('stroke-opacity', '0.05');
        }
    };
</script>

</body>\
'''
toggle_function_template = '''\
javascript:(function(){{
    $all_ranges = $('[id^=length_range]');
    $clicked_ranges = $('[id^=length_range_{outcome_string}]');

    $all_ranges.hide();
    $clicked_ranges.show();

    $all_lines = $('[id^=line_highlighted], [id^=line_nonhighlighted]').find('path');

    $all_lines.css({{
        'stroke-width': '1',
        'stroke-opacity': '0.3'
    }});

    /* For lines with initial opacity below 0.3, keep them below 0.3. */
    $('[id^=line_nonhighlighted_05]').find('path').css('stroke-opacity', '0.05');

    $clicked_lines = $('[id^=line_highlighted_{outcome_string}]').find('path');
    $clicked_lines.css({{
        'stroke-width': '2',
        'stroke-opacity': '1'
    }});

    $('[id^=outcome_category]').find('path').css('stroke-opacity', '0.5');
    $('[id^=outcome_{outcome_string}]').find('path').css('stroke-opacity', '1');

}} )();'''.format

toggle_zoom_function_template = '''\
javascript:(function(){{
    /* axes are id'd starting with axes_1 */

    $next_ax = $('#axes_' + ({panel_i} + 1 + 1));
    $next_lines = $('[id^=zoom_dotted_line_' + {panel_i} + ']');

    if ($next_ax.css('visibility') == 'hidden') {{
        $next_ax.css('visibility', 'visible');
        $next_lines.css('visibility', 'visible');
    }} else {{
        $next_ax.css('visibility', 'hidden');
        $next_lines.css('visibility', 'hidden');
    }};

    for (i = {panel_i} + 1 + 1; i < {num_panels}; i++) {{
        $ax = $('#axes_' + (i + 1));
        $lines = $('[id^=zoom_dotted_line_' + (i - 1) + ']');
        $ax.css('visibility', 'hidden');
        $lines.css('visibility', 'hidden');
    }};
}} )();'''.format

def decorate_outcome_browser(exp):
    fig = exp.plot_outcome_stratified_lengths()
    num_panels = len(fig.axes)

    # Write matplotlib svg to a string.
    with io.StringIO() as write_fh:
        fig.savefig(write_fh, format='svg', bbox_inches='tight')
        contents = write_fh.getvalue()

    plt.close(fig)    

    # Read matplotib svg into an ElementTree.
    with io.StringIO(contents) as read_fh:
        d = ET.parse(read_fh)

    # ElementTree's handling of namespaces is confusing.
    # Learned this approach from http://blog.tomhennigan.co.uk/post/46945128556/elementtree-and-xmlns
    # but don't really understand it.

    default_namespace = 'http://www.w3.org/2000/svg'
    xlink_namespace = 'http://www.w3.org/1999/xlink'
    namespaces = {
        '': default_namespace,
        'xlink': xlink_namespace,
    }

    for prefix, uri in namespaces.items():
        ET.register_namespace(prefix, uri)

    elements_to_decorate = {
        'length_range': [],
        'outcome': [],
        'zoom_toggle_top': [],
    }

    for element in d.iter(f'{{{default_namespace}}}g'):
        if 'id' in element.attrib:
            match = re.match('axes_(?P<panel_i_plus_one>\d+)', element.attrib['id'])
            if match:
                panel_i = int(match.group('panel_i_plus_one')) - 1
                if panel_i > 0:
                    element.attrib['style'] = 'visibility: hidden'
            
            match = re.match('zoom_dotted_line', element.attrib['id'])
            if match:
                element.attrib['style'] = 'visibility: hidden'

            for prefix in elements_to_decorate:
                if element.attrib['id'].startswith(prefix):
                    elements_to_decorate[prefix].append(element)

    def decorate_with_popover(group, container_selector='body', inline_images=False):
        path = group.find(f'{{{default_namespace}}}path', namespaces)
        path.attrib.update({
            'style': 'fill:#000000;stroke:#000000;stroke-linejoin:miter;',
            'fill-opacity': '0.00',
            'stroke-opacity': '0.00',
        })
        
        match = re.match('length_range_(?P<outcome>.+)_(?P<start>\d+)_(?P<end>\d+)', group.attrib['id'])
        
        sanitized_outcome, start, end = match.groups()

        if sanitized_outcome == 'all':
            fns = exp.fns
        else:
            outcome = layout.sanitized_string_to_outcome(sanitized_outcome)
            fns = exp.outcome_fns(outcome)

        fn = fns['length_range_figure'](start, end)

        if fn.exists() or not inline_images:
            if inline_images:
                URI, width, height = table.fn_to_URI(fn)
            else:
                relative_path = fn.relative_to(exp.fns['dir'])
                URI = str(relative_path)
                if fn.exists():
                    with PIL.Image.open(fn) as im:
                        width, height = im.size
                else:
                    width, height = 100, 100

            width = width * 0.75
            height = height * 0.75

            attrib = {
                'data-toggle': 'popover',
                'data-container': container_selector,
                'data-trigger': 'hover click',
                'data-html': 'true',
                'data-placement': 'bottom',
                'data-content': f'<img width={width} height={height} src={URI}>',
            }

            decorator = ET.Element(f'{{{default_namespace}}}a', attrib=attrib)
            decorator.append(path)
            group.remove(path)
            group.append(decorator)

    def decorate_with_toggle(group):
        path = group.find(f'{{{default_namespace}}}path', namespaces)
        
        match = re.match('outcome_(?P<outcome>.+)', group.attrib['id'])
        outcome_string = match.group('outcome')
        
        attrib = {
            f'{{{xlink_namespace}}}href': toggle_function_template(outcome_string=outcome_string),
        }
        
        decorator = ET.Element(f'{{{default_namespace}}}a', attrib=attrib)
        decorator.append(path)
        group.remove(path)
        group.append(decorator)
    
    def decorate_with_zoom_toggle(group):
        path = group.find(f'{{{default_namespace}}}path', namespaces)
        
        match = re.match('zoom_toggle_top_(?P<panel_i>.+)', group.attrib['id'])
        panel_i = match.group('panel_i')
        
        attrib = {
            f'{{{xlink_namespace}}}href': toggle_zoom_function_template(panel_i=panel_i, num_panels=num_panels),
        }
        
        decorator = ET.Element(f'{{{default_namespace}}}a', attrib=attrib)
        decorator.append(path)
        group.remove(path)
        group.append(decorator)

    for group in elements_to_decorate['length_range']:
        decorate_with_popover(group)
        
    for group in elements_to_decorate['outcome']:
        decorate_with_toggle(group)
    
    for group in elements_to_decorate['zoom_toggle_top']:
        decorate_with_zoom_toggle(group)

    with exp.fns['outcome_browser'].open('w') as fh:
        fh.write(before_svg_template(title=exp.name))
        d.write(fh, encoding='unicode')
        fh.write(after_svg)

def length_plot_with_popovers(exp, outcome=None, standalone=False, container_selector='body', x_lims=None, y_lims=None, inline_images=True):
    fig = exp.length_distribution_figure(outcome=outcome, show_ranges=True, x_lims=x_lims, y_lims=y_lims)

    with io.StringIO() as buf:
        fig.savefig(buf, format='svg', bbox_inches='tight')
        # skip matplotlib header lines
        lines = buf.getvalue().splitlines(keepends=True)[4:]

    plt.close(fig)

    length_range_pattern = re.compile('<g id="length_range_(?P<outcome>.+)_(?P<start>\d+)_(?P<end>\d+)"')
    output_lines = []

    if outcome is None:
        title = exp.name
    else:
        category, subcategory = outcome
        title = f'{category}: {subcategory}'
    if standalone:
        output_lines.append(before_svg.format(title=title))

    if outcome is None:
        fns = exp.fns
    else:
        fns = exp.outcome_fns(outcome)

    line_i = 0
    while line_i < len(lines):
        line = lines[line_i]
        output_lines.append(line)

        match = re.search(length_range_pattern, line)
        if match:
            outcome, start, end = match.groups()
            fn = fns['length_range_figures'] / f'{start}_{end}.png'
            if fn.exists():
                if inline_images:
                    URI, width, height = table.fn_to_URI(fn)
                else:
                    relative_path = fn.relative_to(fns['dir'])
                    URI = str(relative_path)
                    with PIL.Image.open(fn) as im:
                        width, height = im.size

                before_path = before_path_template.format(URI=URI,
                                                          width=width,
                                                          height=height,
                                                          container_selector=container_selector,
                                                         )
                output_lines.append(before_path)
                line_i += 1
                line = lines[line_i]
                while not line.endswith('/>\n'):
                    output_lines.append(line)
                    line_i += 1
                    try:
                        line = lines[line_i]
                    except IndexError:
                        print(line_i, len(lines))
                        print(lines[-1])
                        raise

                output_lines.append(after_path)
        line_i += 1

    if standalone:
        output_lines.append(after_svg)

    return '\n'.join(output_lines)
